/* $XConsortium: miZBuffer.c,v 1.7 94/01/07 15:06:33 takamori Exp $ */

#include "gcstruct.h"
#include "ddpex3.h"
#include "stdio.h"

#define _DEBUG_ON_
#define _BOUNDS_CONTROL_ON_

/*++
 |
 |  Function Name:	ZBuffer_allocate
 |
 |  Function Description:
 |	 Allocates the storage needed for the ZBuffer structure
 |
 |  Note(s):
 |
 --*/

ddpex3rtn
ZBuffer_allocate(Pzbs_ptr, drawable_ptr)
/* in */ 
     DrawablePtr drawable_ptr;
/* out */
     ddZBufferStruct **Pzbs_ptr;
{
    /*
     * Allocate storage for the z-buffer structure, and initialize its size.
     *  Then allocate and initialize the actual buffer.  If any error occurs,
     *  then deallocate the structure.
     */
    if (!(*Pzbs_ptr = (ddZBufferStruct *) Xalloc(sizeof(ddZBufferStruct))))
        return (BadAlloc);

    (*Pzbs_ptr)->xsize = drawable_ptr->width;
    (*Pzbs_ptr)->ysize = drawable_ptr->height;
    if (!((*Pzbs_ptr)->buffer = (ddZBuffer) Xalloc(sizeof(ddZBufferEntry)
					       * (*Pzbs_ptr)->xsize 
					       * (*Pzbs_ptr)->ysize))) {
        Xfree(*Pzbs_ptr);
	return (BadAlloc);
    }
    ZBuffer_clear(*Pzbs_ptr);
    return (Success);
}				/* ZBuffer_allocate */

/*++
 |
 |  Function Name:	ZBuffer_free
 |
 |  Function Description:
 |	 Frees up the storage used by the ZBuffer structure
 |
 |  Note(s):
 |
 --*/

ddpex3rtn
ZBuffer_free(zbs_ptr)
/* in */ 
/* out */
     ddZBufferStruct *zbs_ptr;
{
    if (zbs_ptr) {
        if (zbs_ptr->buffer)
	    Xfree(zbs_ptr->buffer);
	Xfree(zbs_ptr);
	zbs_ptr = NULL;
    }
    return (Success);
}				/* ZBuffer_free */

/*++
 |
 |  Function Name:	ZBuffer_clear
 |
 |  Function Description:
 |	 Clears the Z-Buffer.
 |
 |  Note(s):
 |
 --*/

ddpex3rtn
ZBuffer_clear(zbs_ptr)
/* in */ 
     ddZBufferStruct *zbs_ptr;
/* out */
{
    int i, bytes, buffsize;
    
    buffsize = zbs_ptr->xsize * zbs_ptr->ysize;
    bzero(zbs_ptr->buffer, buffsize*sizeof(ddZBufferEntry));
    return (Success);
}				/* ZBuffer_clear */


/*++
 |
 |  Function Name:	ZBuffer_FillSpans
 |
 |  Function Description:
 |	 Fills multiple spans in the Z-Buffer.
 |
 |  Note(s):
 |
 --*/

ddpex3rtn
ZBuffer_FillSpans(zbs_ptr, dst, pgc, nPts, pnts, widths, depths)
/* in */ 
     ddZBufferStruct *zbs_ptr;
     DrawablePtr     dst;
     GCPtr           pgc;
     int             nPts;
     ddCoord3DS      *pnts;
     int             *widths;
     int             *depths;
/* out */
{
    int x, y, z;
    int x2, dx, dz;
    int m, m1, incr1, incr2, d;
    int offset, bufforg;
    int draw, odraw;
    int i, outnPts, *outwidths;
    ddCoord2DS *outpnts;

#ifdef _DEBUG_ON_
int times, index;
times=0; index = 0;
#endif
    if (pgc->miTranslate)
        bufforg = dst->y * zbs_ptr->xsize + dst->x;
    else
        bufforg = 0;
    outnPts = 0;
    outpnts = (ddCoord2DS *) Xalloc(sizeof(ddCoord2DS) * nPts);
    outwidths = (int *) Xalloc(sizeof(int) * nPts);
    if (!outwidths || !outpnts)
    {
        if (outpnts) Xfree(outpnts);
        if (outwidths) Xfree(outwidths);
	return (BadAlloc);
    }
    for (i = 0; i < nPts; i++) {
	x = x2 = pnts->x;
	y = pnts->y;
#ifdef _BOUNDS_CONTROL_ON_
	if ((y >= dst->y) && (y <= zbs_ptr->ysize + dst->y)) {
#endif
	  z = pnts->z;
	  offset = y * zbs_ptr->xsize + x - bufforg;
	  dz = *depths;
	  dx = *widths;
	  if (dx == 0) dx = 1;
	  /* This bresenham-like code was lifted from the polygon-filling
	     procedures.. It works like Bresenham's line-drawing algorithm.
	     I have modified it in order to work in a 3-d span fashion instead 
	     of a 2-d line-drawing method. This means that it works in the x-z 
	     plane, instead of the x-y plane.  */
	  m = dz / dx;
	  if (dz < 0) {
	      m1 = m - 1;
	      incr1 = -2 * dz + 2 * dx * m1;
	      incr2 = -2 * dz + 2 * dx * m;
	      d = 2 * m * dx - 2 * dz - 2 * dx;
	  } else {
	      m1 = m + 1;
	      incr1 = 2 * dz - 2 * dx * m1;
	      incr2 = 2 * dz - 2 * dx * m;
	      d = -2 * m * dx + 2 * dz;
	  }
	  /* We do a bresenham's across the x-z plane at a fixed y value.  The
	     spans are generated at each point where the drawing should stop
	     (we use the zbuffer to determine if a pixel should be drawn) */
	  if (odraw = draw = (zbs_ptr->buffer[offset] <= z))
	    zbs_ptr->buffer[offset] = z;
#ifdef _BOUNDS_CONTROL_ON_
	  while (x2 < pnts->x + dx && x2 < zbs_ptr->xsize + dst->x) {
	      while (draw == odraw
		     && x2 < pnts->x + dx && x2 < zbs_ptr->xsize + dst->x) {
#else
	  while (x2 < pnts->x + dx) {
	      while (draw == odraw && x2 < pnts->x + dx) {
#endif
		  x2++;
		  if (m1 > 0) {
	              if (d > 0) {
		          z += m1;
			  d += incr1;
		      } else {
		          z += m;
			  d += incr2;
		      }
		  } else {
	              if (d >= 0) {
		          z += m1;
			  d += incr1;
		      } else {
		          z += m;
			  d += incr2;
		      }
		  }
		  offset++;
		  if (draw = zbs_ptr->buffer[offset] <= z)
		      zbs_ptr->buffer[offset] = z;
	      }
	      if (odraw) {
	          /* if the output buffer is full then flush it, otherwise just
		     generate the span */
	          if (outnPts == nPts) {
		      (*pgc->ops->FillSpans) (dst, pgc, outnPts, outpnts, 
					      outwidths, TRUE);
		      outnPts = 0;
		  }
#ifdef _DEBUG_ON_
if ((x < dst->x) || (x2 > zbs_ptr->xsize + dst->x)
    || (y < dst->y) || (y > zbs_ptr->ysize + dst->y)) {
    x /= (y - y);
  }
times++;
#endif
		  outpnts[outnPts].x = x;
		  outpnts[outnPts].y = y;
		  outwidths[outnPts] = x2 - x;
		  outnPts++;
	      }
	      x = x2;
	      odraw = draw;
#ifdef _BOUNDS_CONTROL_ON_
	  }
#endif
	}  /* Next zbuffer span */
#ifdef _DEBUG_ON_
index++;
#endif
	pnts++;
	depths++;
	widths++;
      }
    /* flush any spans left over to be drawn */
    (*pgc->ops->FillSpans) (dst, pgc, outnPts, outpnts, outwidths, TRUE);
    Xfree(outpnts);
    Xfree(outwidths);
    return (Success);
}				/* ZBuffer_spanfill */
